@@ -56,10 +56,44 @@ import copy, warnings
 
 
 class CEM:
-    def __init__(self, phi0, batch_size=0, ref_mode=None, ref_thresh=None,
-                 ref_alpha=0.05, n_orig_per_batch=None, min_batch_update=0.2,
-                 force_min_samples=True, soft_update=0, soft_q_update=0,
-                 w_clip=0, optim_mode=False, title='CEM'):
+    def __init__(
+            self,
+            # initial distribution parameters
+            phi0,
+
+            #########   General   #########
+            # number of samples between distribution updates
+            # (0 = no updates)
+            batch_size=0,
+            # optimization mode: force ref_mode='none' and w_clip=1
+            optim_mode=False,
+            # clip IS weights to the range (1/w_clip, w_clip)
+            # (0 = no clipping; 1 = no weights; 0<w_clip<1 will be inverted)
+            w_clip=0,
+            # experiment title
+            title='CEM',
+
+            #########   Sampling parameters   #########
+            # number of reference samples per batch, taken from the original distribution
+            n_orig_per_batch=None,
+
+            #########   Update-step parameters   #########
+            # where to estimate the reference-distribution from (see details below)
+            ref_mode=None,
+            # *absolute* objective: aim to sample X<=ref_thresh
+            ref_thresh=None,
+            # *quantile* objective (if ref_thresh==None): aim to sample ref_alpha-tail
+            ref_alpha=0.05,
+            # minimum samples to use for update step
+            # (fraction for percent; int for absolute)
+            min_batch_update=0.2,
+            # when filtering R<q, add samples with R==q if update-samples are too few
+            force_min_samples=True,
+            # phi <- (1-soft_update)*new_phi + soft_update*current_phi
+            soft_update=0,
+            # quantile_estimate <- (1-soft_q_update)*estimate + soft_q_update*prev_estimate
+            soft_q_update=0,
+    ):
         self.title = title
         self.default_dist_titles = None
         self.default_samp_titles = None
@@ -212,7 +246,7 @@ class CEM:
             self.batch_count, self.sample_count, self.update_count, self.ref_scores,
             self.sample_dist, self.sampled_data, self.weights, self.is_reference,
             self.scores, self.ref_quantile, self.internal_quantile, self.selected_samples,
-            self.n_update_samples
+            self.n_update_samples, self.optim_mode
         )
         with open(filename, 'wb') as h:
             pkl.dump(obj, h)
@@ -227,7 +261,7 @@ class CEM:
         self.batch_count, self.sample_count, self.update_count, self.ref_scores, \
         self.sample_dist, self.sampled_data, self.weights, self.is_reference, \
         self.scores, self.ref_quantile, self.internal_quantile, self.selected_samples, \
-        self.n_update_samples = obj
+        self.n_update_samples, self.optim_mode = obj
 
     def is_original_dist(self, shuffle=True):
         if not shuffle:
@@ -241,6 +275,10 @@ class CEM:
         dist = self.sample_dist[0] if orig_dist else self.sample_dist[-1]
         x = self.do_sample(dist)
         w = self.get_weight(x, orig_dist)
+        self.update_sample(x, w, orig_dist)
+        return x, w
+
+    def update_sample(self, x, w, orig_dist):
         self.sampled_data[-1].append(x)
         self.weights[-1].append(w)
         self.is_reference[-1].append(orig_dist)
@@ -249,7 +287,6 @@ class CEM:
             warnings.warn(f'Drawn {self.sample_count}>{self.batch_size} samples '
                           f'without updating (only {self.update_count}<'
                           f'{self.batch_size} scores for update)')
-        return x, w
 
     def sample_batch(self, n=None, shuffle=True):
         if n is None: n = self.batch_size
@@ -455,7 +492,7 @@ class CEM:
             score=np.concatenate(s),
         )
         for k,v in samples.items():
-            d2_dict[k] = v
+            d2_dict[k] = v[:len(d2_dict['weight'])]
         d2 = pd.DataFrame(d2_dict)
 
         return d1, d2
@@ -501,6 +538,9 @@ class CEM:
                         lambda d: cvar(d.score.values,self.ref_alpha))
                     ax.plot(cvar_orig, label=f'Reference {cvar_lab:s}')
             else:
+                if self.w_clip > 0:
+                    warnings.warn('Reference distribution reconstruction may be inaccurate '
+                                  'with clipped weights and without reference samples.')
                 mean_orig = c2.groupby('batch').apply(
                     lambda d: np.mean(d.score*d.weight)/d.weight.mean())
                 ax.plot(mean_orig, label='Reference mean (IS)')
@@ -522,6 +562,68 @@ class CEM:
         ax.legend(fontsize=13)
         return ax
 
+    def show_tail_level(self, ax=None):
+        if ax is None:
+            ax = plt.subplots(1, 1, figsize=(5,3.5))[1]
+            ax.grid(color='k', linestyle=':', linewidth=0.3)
+
+        c1, c2 = self.get_data()
+
+        if self.n_orig_per_batch > 0:
+            tail_level = []
+            for b in range(c2.batch.values[-1]+1):
+                bb = c2[c2.batch==b]
+                sample_mean = bb[~bb.is_ref].score.mean()
+                refs = sorted(bb[bb.is_ref].score.values)
+                ref_means = np.cumsum(refs) / np.arange(1,len(refs)+1)
+                alphas = np.where(ref_means<=sample_mean)[0]
+                alpha = 0 if len(alphas)==0 else 100 * alphas[-1] / len(ref_means)
+                tail_level.append(alpha)
+
+        else:
+            if self.w_clip > 0:
+                warnings.warn('Reference distribution reconstruction may be inaccurate '
+                              'with clipped weights and without reference samples.')
+
+            def wcvar(x, w, alpha):
+                q = quantile(x, alpha, w)
+                ids = x <= q
+                x = x[ids]
+                w = w[ids]
+                return np.mean(x * w) / np.mean(w)
+
+            tail_level = []
+            for b in range(c2.batch.values[-1]+1):
+                bb = c2[c2.batch==b]
+                sample_mean = bb[~bb.is_ref].score.mean()
+                x, w = bb.score.values, bb.weight.values
+                refs = np.array([wcvar(x,w,alpha) for alpha in np.linspace(0,1,101)])
+                alpha = np.where(refs<=sample_mean)[0][-1]
+                tail_level.append(alpha)
+
+        if self.ref_thresh is None and not self.optim_mode:
+            ax.axhline(100*self.ref_alpha, color='k', linestyle='--')
+        ax.plot(tail_level)
+        ax.set_ylim((-1, 101))
+        ax.set_xlabel('iteration', fontsize=15)
+        ax.set_ylabel('tail level [%]', fontsize=15)
+        ax.set_title(r'$\alpha$ | $mean$(sample)==$CVaR_{\alpha}$(ref)', fontsize=15)
+        return ax
+
+    def show_summary(self, axs=None, ylab=None):
+        if self.optim_mode:
+            return self.show_sampled_scores(axs, ylab)
+
+        if axs is None:
+            axs = plt.subplots(1, 2, figsize=(10,4))[1]
+            for i in range(2):
+                axs[i].grid(color='k', linestyle=':', linewidth=0.3)
+
+        self.show_sampled_scores(axs[0], ylab)
+        self.show_tail_level(axs[1])
+        plt.tight_layout()
+        return axs
+
 
 def quantile(x, q, w=None, is_sorted=False, estimate_underlying_quantile=False):
     n = len(x)
